/** @file emu_screen.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2017.12.01

	@brief [ emu screen ]
*/

#include "../emu.h"
#include "../vm/vm.h"
#include "../gui/gui.h"
#include "../config.h"
#ifdef USE_MESSAGE_BOARD
#include "../msgboard.h"
#endif
#include "../csurface.h"
#include "../video/rec_video.h"
#include "../utils.h"

static const VmSize window_proposal[] = {
	{640,400},{640,480},{800,600},{1024,768},{0,0}
};

void EMU::release_screen_on_emu_thread()
{
	// stop video recording
	stop_rec_video();
}

#if 0
int EMU::get_window_width(int mode)
{
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_height + screen_height_aspect * mode;
	}
#endif
	return window_width + screen_width_aspect * mode;
}

int EMU::get_window_height(int mode)
{
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_width + screen_width_aspect * mode;
	}
#endif
	return window_height + screen_height_aspect * mode;
}
#endif

/// set vm drawing area
void EMU::set_vm_display_size()
{
	if (vm) {
		int new_size = config.screen_video_size;
		int l = mixed_dest_x;
		int t = mixed_dest_y;
		int r = l + mixed_width;
		int b = t + mixed_height;

		// vm drawing area is same as area to capture and record screen
		if (rec_video_size[new_size].w > mixed_width) {
			l = rec_video_size[new_size].x;
			r = rec_video_size[new_size].x + rec_video_size[new_size].w;
		}

		if (rec_video_size[new_size].h > mixed_height) {
			t = rec_video_size[new_size].y;
			b = rec_video_size[new_size].y + rec_video_size[new_size].h;
		}

		vm->set_display_size(l,t,r,b);
	}
}

/// reserve top margin area for ledbox
int EMU::adjust_y_position(int h, int y)
{
	if (20 < h - LIMIT_MIN_WINDOW_HEIGHT && h - LIMIT_MIN_WINDOW_HEIGHT < 42) {
		y -= (42 - h + LIMIT_MIN_WINDOW_HEIGHT) / 2;
	}
	return y;
}

void EMU::fill_gray()
{
	scrntype color = map_rgbcolor(0x40, 0x40, 0x40);
	for(int y=0; y<source_height; y++) {
		scrntype *line = screen_buffer(y);
		for(int x=0; x<source_width; x++) {
			*line = color;
			line++;
		}
	}
}

bool EMU::now_skip_frame()
{
	return skip_frame && !now_rec_video();
}

///
/// stop recording video
///
void EMU::stop_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Stop();
#endif
}

///
/// restart recording video
///
void EMU::restart_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Restart();
#endif
}

bool EMU::now_rec_video()
{
#ifdef USE_REC_VIDEO
	return *now_recording_video;
#else
	return false;
#endif
}

void EMU::resize_rec_video(int num)
{
#ifdef USE_EMU_INHERENT_SPEC
	config.screen_video_size = num;
	change_rec_video_size(num);

	// send display size to vm
	set_vm_display_size();
#endif
}
void EMU::change_rec_video_size(int num)
{
	rec_video_stretched_size = rec_video_size[num];
	if (mixed_width_ratio < mixed_height_ratio) {
		rec_video_stretched_size.h = rec_video_stretched_size.h * mixed_width_ratio / mixed_height_ratio;
		rec_video_stretched_size.y += (rec_video_size[num].h - rec_video_stretched_size.h) / 2; 
	} else {
		rec_video_stretched_size.w = rec_video_stretched_size.w * mixed_height_ratio / mixed_width_ratio;
		rec_video_stretched_size.x += (rec_video_size[num].w - rec_video_stretched_size.w) / 2; 
	}
	rec_video_stretched_size.y = adjust_y_position(rec_video_stretched_size.h, rec_video_stretched_size.y);

#ifdef _DEBUG_LOG
	out_debugf(_T("          rec_video_size: num:%d x:%d y:%d w:%d h:%d")
		,num,rec_video_size[num].x,rec_video_size[num].y,rec_video_size[num].w,rec_video_size[num].h);
	out_debugf(_T("rec_video_stretched_size:       x:%d y:%d w:%d h:%d")
		,rec_video_stretched_size.x,rec_video_stretched_size.y,rec_video_stretched_size.w,rec_video_stretched_size.h);
#endif
}

bool EMU::rec_video_enabled(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->IsEnabled(type);
#else
	return false;
#endif
}

const _TCHAR **EMU::get_rec_video_codec_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetCodecList(type);
#else
	return NULL;
#endif
}

const _TCHAR **EMU::get_rec_video_quality_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetQualityList(type);
#else
	return NULL;
#endif
}

//----------

int EMU::qsort_by_w_h_n(const void *a, const void *b)
{
	struct window_mode_st *a1 = (struct window_mode_st *)a;
	struct window_mode_st *b1 = (struct window_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	if (!rc) rc = (a1->power - b1->power);
	return rc;
}

int EMU::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	int aa = a1->width * a1->height;
	int bb = b1->width * b1->height;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (bb - aa);
	return rc;
}

/// enumerate screen mode for window
/// calcurate magnify range
/// @param [in] max_width
/// @param [in] max_height
void EMU::enum_window_mode(int max_width, int max_height)
{
	// enumerate screen mode for window
	int w,h;
	bool minsize = false;
	bool maxsize = false;

	window_mode_count = 0;
	for (int i=0; window_proposal[i].cx != 0 && window_mode_count < WINDOW_MODE_MAX; i++) {
		w = window_proposal[i].cx;
		h = window_proposal[i].cy;
		// minimum window size?
		if (w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT) {
			continue;
		}
		if (!minsize) {
			w = MIN_WINDOW_WIDTH;
			h = MIN_WINDOW_HEIGHT;
			i--;
			minsize = true;
		}
		if (w > MAX_WINDOW_WIDTH || h > MAX_WINDOW_HEIGHT) {
			if (!maxsize) {
				w = MAX_WINDOW_WIDTH;
				h = MAX_WINDOW_HEIGHT;
				maxsize = true;
				i--;
			} else {
				break;
			}
		}
		for (int pow = 10; pow <= 30 && window_mode_count < WINDOW_MODE_MAX; pow+=5) {
			if ((w*pow/10) <= max_width && (h*pow/10) <= max_height) {
				bool found = false;
				for(int j = 0; j < window_mode_count; j++) {
					if(window_mode[j].power == pow && window_mode[j].width == (w*pow/10) && window_mode[j].height == (h*pow/10)) {
						found = true;
						break;
					}
				}
				if (!found) {
					window_mode[window_mode_count].power = pow;
					window_mode[window_mode_count].width = w * pow / 10;
					window_mode[window_mode_count].height = h * pow / 10;
					out_debugf(_T("window_mode:%d %dx%d x%.1f"),window_mode_count, w * pow / 10, h * pow / 10, (double)pow / 10.0);
					window_mode_count++;
				}
			} else {
				break;
			}
		}
	}
	qsort(&window_mode, window_mode_count, sizeof(struct window_mode_st), qsort_by_w_h_n);
}


/// matching window or fullscreen size
void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		// matching window size at last time
		bool find = false;
		int i = config.window_mode;
		if (i < window_mode_count) {
			if (window_mode[i].width == config.screen_width
			&&  window_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	} else {
		// matching fullscreen size at last time
		bool find = false;
		for(int i = 0; i < screen_mode_count; i++) {
			if (screen_mode[i].disp_no == config.disp_device_no
			&&  screen_mode[i].width == config.screen_width
			&&  screen_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i + 8;
				break;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	}
}

void EMU::change_stretch_screen(int num)
{
	if (num < 0) {
		num = (config.stretch_screen + 1);
	}
	num = (num % 3);
	if (config.stretch_screen == num) {
		num = 0;
	}
	config.stretch_screen = num;
	if (now_fullscreen) {
		set_display_size(-1, -1, 10, !now_fullscreen);
	}
}

#if 0
void EMU::change_cutout_screen()
{
	config.cutout_screen = !config.cutout_screen;
	if (now_fullscreen) set_display_size(-1, -1, 10, !now_fullscreen);
}
#endif

void EMU::change_pixel_aspect(int mode)
{
	if (mode < 0) {
		mode = (config.pixel_aspect + 1) % get_pixel_aspect_count();
	}
	config.pixel_aspect = get_pixel_aspect(mode, &mixed_width_ratio, &mixed_height_ratio);

	set_display_size(config.screen_width, config.screen_height, now_fullscreen ? 10 : window_mode_power, !now_fullscreen);
}

int EMU::get_pixel_aspect_count()
{
	return 3;
}

int EMU::get_pixel_aspect(int mode, int *wratio, int *hratio)
{
	int new_mode = mode;
	switch(mode) {
	case 1:
		*wratio = 10;
		*hratio = 11;
		break;
	case 2:
		*wratio = 10;
		*hratio = 12;
		break;
	default:
		*wratio = 1;
		*hratio = 1;
		new_mode = 0;
		break;
	}
	return new_mode;
}

window_mode_t *EMU::get_window_mode(int num)
{
	if (num < window_mode_count) {
		return &window_mode[num];
	}
	return NULL;
}

screen_mode_t *EMU::get_screen_mode(int num)
{
	if (num < screen_mode_count) {
		return &screen_mode[num];
	}
	return NULL;
}

VmRectWH *EMU::get_screen_record_size(int num)
{
	if (0 <= num && num <= 1) {
		return &rec_video_size[num];
	}
	return NULL;
}

void EMU::lock_screen()
{
	mux_update_screen->lock();
	screen_changing = true;
}

void EMU::unlock_screen()
{
	screen_changing = false;
	mux_update_screen->unlock();
}

